#![feature(duration_constructors)]

use mail_parser::{Message};
use std::error::Error;
use std::path::{PathBuf};
use models::*;
use std::ffi::{OsStr, OsString};
use std::{fs};
use std::collections::HashMap;
use std::fs::{create_dir_all, File};
use std::future::Future;
use std::io::{Write};
use std::sync::Mutex;
use std::time::{Instant};
use anyhow::anyhow;
use serde::{Deserialize, Serialize};
use crate::indexes::{Indexes};
use config::{Config, INSTANCE};
use crate::server::{run_api};
use mailparse::{DispositionType, parse_mail, ParsedMail};
use once_cell::sync::Lazy;
use regex::Regex;
use tokio::task::JoinHandle;
use crate::templates::util::parse_email;
use crate::util::{compress_and_save_file};
use crate::imap::{delete_folder, rename_folder, create_folder, check_for_updates, read_local_uids};
use crate::js::email_scripts;
use warp::Filter;
use crate::search::{ingest_new_message, MY_TASK};

mod smtp_client;
mod arg;
mod config;
mod models;
mod templates;
mod util;
mod indexes;
mod imap;
mod server;
mod js;
mod sonic;
mod api_broadcast_structure;
mod search;

struct ImapTaskStorage {
    downloading: Option<JoinHandle<()>>,
    idle_handlers: HashMap<String, JoinHandle<Result<(), anyhow::Error>>>
}

impl ImapTaskStorage {
    fn new() -> Self {
        ImapTaskStorage {
            downloading: None,
            idle_handlers: HashMap::new(),
        }
    }
    
    pub fn add_idle_task(&mut self, mailbox: String, handler: JoinHandle<Result<(), anyhow::Error>>) {
        // TODO check if the task already exists
        // TODO add a task when a new folder spotted on the IMAP
        self.idle_handlers.insert(mailbox, handler);
    }
    
    pub fn remove_idle_task(&mut self, mailbox: String){
        if self.idle_handlers.contains_key(&mailbox.clone()) {
            self.idle_handlers.remove(&mailbox.clone());
        }
    }
    
    pub fn start_downloading(&mut self){
        if let Some(ref mut handle) = self.downloading {
            handle.abort();
        }
        
        self.downloading = Some(tokio::spawn(async move {
            match imap::download_email_from_imap().await {
                Ok(_) => {}
                Err(_) => println!("Cannot download new emails")
            }
        }));
    }
}

pub static IMAP_TASK_STORAGE: Lazy<Mutex<ImapTaskStorage>> = Lazy::new(|| Mutex::new(ImapTaskStorage::new()));


/// appends an extension to a path
pub fn append_ext(ext: impl AsRef<OsStr>, path: &PathBuf) -> PathBuf {
    let mut os_string: OsString = path.into();
    os_string.push(".");
    os_string.push(ext.as_ref());
    os_string.into()
}

/// converts message to HTML string
pub fn message_to_html(msg: StrMessage) -> String{
    fn remove_style_tags(html: &str) -> String {
        Regex::new(r"(?s)<style.*?>.*?</style>").unwrap()
            .replace_all(html, "")
            .to_string()
    }
    fn extract_body(html: &str) -> String {
        let open_tag = "<body";
        let close_tag = "</body>";

        if let Some(start_pos) = html.find(open_tag) {
            if let Some(open_tag_end) = html[start_pos..].find('>') {
                let body_start = start_pos + open_tag_end + 1;
                if let Some(end_pos) = html[body_start..].find(close_tag) {
                    return html[body_start..body_start + end_pos].to_string();
                }
            }
        }
        html.to_string()
    }
    fn extract_styles(html: &str) -> String {
        let style_pattern = Regex::new(r"(?is)<style[^>]*>(.*?)</style>").unwrap();
        style_pattern.captures_iter(html)
            .filter_map(|caps| caps.get(1))
            .map(|m| m.as_str())
            .collect::<Vec<_>>()
            .join("\n")
    }


    let data = match fs::read_to_string(&msg.original_path.clone()) {
        Ok(content) => content,
        Err(e) => {
            eprintln!("Error reading file: {}", e);
            return "".to_string();
        }
    };

    let parsed_mail = match parse_mail(&data.as_bytes()) {
        Ok(mail) => mail,
        Err(e) => return "".to_string(),
    };

    let html = parse_email(&parsed_mail).unwrap_or_else(|_| msg.body.clone());
    let styles = extract_styles(&html);
    let mut body_content = remove_style_tags(&extract_body(&html));
    // body_content = purple_numbers(&*body_content, "#");

    // retrieving attachments
    retrieve_and_save_attachments(parsed_mail, msg.list.clone(), msg.clone().id)
        .expect("Unable to retrieve attachment");
    
    format!(
        r#"
                <!DOCTYPE html>
                <html>
                <head>
                    <title>Email</title>
                    <meta http-equiv='Permissions-Policy' content='interest-cohort=()'/>
                    <meta name='viewport' content='width=device-width, initial-scale=1.0, maximum-scale=1.0,user-scalable=0' />
                </head>   
                {js}   
                <body>         
                    <div id="email-body" class="email-body" style="position: relative;">
                        <style>
                            {styles}
                        </style>
                        <div">
                            {body_content}
                        </div>
                    </div>
                </body>
                </html>
            "#,
        js = email_scripts(),
        styles = styles,
        body_content = body_content,
    )
}

fn retrieve_and_save_attachments(parsed_mail: ParsedMail, list: String, file_name: String) -> anyhow::Result<()>{
    // iterating through each part of email
    for part in parsed_mail.subparts {
        let disposition = part.get_content_disposition();
        // if this is an attachment
        if disposition.disposition == DispositionType::Attachment {
            // creating directory for it
            let full_dir = Config::global().out_dir.clone()
                .join(list.clone())
                .join(".attachments")
                .join(file_name.clone());
            create_dir_all(full_dir.clone()).expect("Unable to create directory for attachments");

            let mut filename = disposition.params.get("filename")
                .map(|s| s.to_string())
                .unwrap_or_else(|| "unnamed_file".to_string());

            // creating full absolute path to the file
            let full_path = full_dir.join(filename.clone());
            let mut output = File::create(full_path.clone())?;

            // writing the data to the file
            output.write_all(&part.get_body_raw().unwrap())?;
        }
    }

    Ok(())
}

/// saves email to local file system
fn persist_email(msg: &Message, uid: u32, list: String, original_path: PathBuf) -> anyhow::Result<()>{
    // creating all folders
    if Config::global().out_dir.exists() {
        std::fs::create_dir_all(&Config::global().out_dir).ok();
    }
    let message_dir = Config::global().out_dir.join(list.clone()).join("messages");
    if !message_dir.exists() {
        std::fs::create_dir_all(&message_dir).ok();
    }

    // persist email
    let mut message = StrMessage::new(msg, uid, list.clone());
    message.set_original_path(original_path);
    let mut hash_path = message_dir.clone().join(message.hash.clone());
    hash_path = append_ext("html", &hash_path.clone());
    let relative_path = match hash_path.clone().strip_prefix(Config::global().out_dir.clone()) {
        Ok(relative_path) => {PathBuf::from(relative_path)}
        Err(_) => {hash_path.clone()}
    };
    message.set_path(relative_path.display().to_string());
    if Config::global().include_html {
        let data = message_to_html(message.clone());
        match compress_and_save_file(message_dir.clone(), message.hash.clone(), data.as_bytes()){
            Ok(_) => {}
            Err(_) => {println!("Error while compressing or saving {}", message.hash.clone())}
        };
    }
    
    // ingest new massage to the sonic server
    ingest_new_message(message.clone())?;
    
    // indexing
    let _ = Indexes::add(message.clone());
    
    Ok(())
}

/// read raw email and saves HTML version
pub async fn add_email(path: PathBuf, uid: u32) -> anyhow::Result<()>{
    let maildir = match path.parent() {
        None => {return Err(anyhow!("Cannot retrieve parent folder from the path"))}
        Some(parent) => {
            match parent.parent() {
                None => {return Err(anyhow!("Cannot retrieve parent folder from the path"))}
                Some(parent) => {parent.file_name().unwrap().to_owned().into_string().unwrap()}
            }
        }
    };
    let maildir_out_path = Config::global().out_dir.clone().join(maildir.clone());
    if !maildir_out_path.exists() { create_dir_all(maildir_out_path).ok(); }

    let data = std::fs::read(&path).expect("could not read mail");

    if let Some(mail) = Message::parse(&data) {
        let _ = persist_email(&mail, uid.clone(), maildir.clone(), path.clone());
    } else {
        println!("Could not parse message {:?}", path);
    }
    
    Ok(())
}

/// delete a single message both locally and remotely
async fn delete_email(message_id: String) -> anyhow::Result<()>{
    let message = Indexes::delete_from_messages(message_id.clone());
    match message {
        Ok(message) => {
            // delete from indexes
            let _ = Indexes::delete_from_threads(message.clone());
            let _ = Indexes::delete_from_messages(message_id.clone());
            let lists: Vec<String> = fs::read_dir(Config::global().out_dir.clone())?
                .filter_map(Result::ok)
                .filter(|entry| entry.metadata().map(|m| m.is_dir()).unwrap_or(false))
                .filter_map(|entry| entry.file_name().into_string().ok())
                .collect();
            let _ = Indexes::persist_indexes(lists);

            // delete HTML file
            let message_html_name = Config::global().out_dir.clone()
                .join(message.list.clone())
                .join("messages")
                .join(message.hash.clone() + ".html");
            let _ = fs::remove_file(message_html_name);

            // delete uid from downloaded uid list
            let uids_path = Config::global().maildir.clone().join(".uids.json");
            let mut uids_local = read_local_uids(uids_path.clone()).unwrap();
            if let Some(set) = uids_local.get_mut(&message.list.clone()) {
                set.remove(&message.uid.clone().to_string());
            }

            let _ = imap::delete_email_by_uid(message.list.clone(), message.uid.clone()).await;
        }
        Err(_) => {}
    }

    Ok(())
}

/// create folder both locally and remotely
pub async fn create_folder_lar(name: String) -> anyhow::Result<()>{
    match create_folder(name.clone()).await {
        Ok(_) => {
            let folder_path = Config::global().out_dir.clone().join(name.clone());
            if !folder_path.exists() {
                return match fs::create_dir_all(folder_path) {
                    Ok(_) => Indexes::persist_indexes(vec![name.clone()]),
                    Err(_) => Err(anyhow!("Cannot create folder locally")),
                }
            }

            // start monitoring
            let handler = check_for_updates(name.clone());
            let mut idle_task_storage = IMAP_TASK_STORAGE.lock().unwrap();
            idle_task_storage.add_idle_task(name.clone(), handler);
            
            Ok(())
        }
        Err(_) => Err(anyhow!("Cannot create folder remotely"))
    }
}

/// delete folder both locally and remotely
pub async fn delete_folder_lar(name: String) -> anyhow::Result<()>{
    match delete_folder(name.clone()).await {
        Ok(_) => {
            let folder_path = Config::global().out_dir.clone().join(name.clone());
            if folder_path.exists() {
                return match fs::remove_dir_all(folder_path) {
                    Ok(_) => Ok(()),
                    Err(_) => Err(anyhow!("Cannot delete folder locally")),
                }
            }

            // stop monitoring the folder
            let mut idle_task_storage = IMAP_TASK_STORAGE.lock().unwrap();
            idle_task_storage.remove_idle_task(name.clone());
            
            Ok(())
        }        
        Err(_) => Err(anyhow!("Cannot delete folder remotely"))
    }
}

/// rename folder both locally and remotely
pub async fn rename_folder_lar(old_name: String, new_name: String) -> anyhow::Result<()>{
    match rename_folder(old_name.clone(), new_name.clone()).await {
        Ok(_) => {
            let old_path = Config::global().out_dir.clone().join(old_name.clone());
            let new_path = Config::global().out_dir.clone().join(new_name.clone());
            if old_path.exists() {
                return match fs::rename(old_path, new_path) {
                    Ok(_) => Ok(()),
                    Err(_) => Err(anyhow!("Cannot rename folder locally"))
                }
            }

            // stop monitoring old folder
            let mut idle_task_storage = IMAP_TASK_STORAGE.lock().unwrap();
            idle_task_storage.remove_idle_task(old_name.clone());
            
            // start monitoring new folder
            let handler = check_for_updates(new_name.clone());
            idle_task_storage.add_idle_task(new_name.clone(), handler);
            
            Ok(())
        }
        Err(_) => Err(anyhow!("Cannot rename folder remotely"))
    }
}

/// Parse args from CLI
fn parse_args(){
    let args = arg::Args::from_env();
    
    let mut config = Config::from_file(&args.config).expect("could not read config");
    if !args.include_html { config.include_html = true; } 
    else { config.include_html = args.include_html; }
    
    config.maildir = args.maildir.clone();
    config.out_dir = args.out_dir.clone();
    INSTANCE.set(config).unwrap();
}

pub fn start_checkin_for_all_updates(){
    // start updates monitoring
    for mailbox in Indexes::get_local_mailboxes() {
        let handler = check_for_updates(mailbox.clone());
        let mut idle_task_storage = IMAP_TASK_STORAGE.lock().unwrap();
        idle_task_storage.add_idle_task(mailbox.clone(), handler);
        drop(idle_task_storage);
    }
} 

async fn run(){
    // parse args from CLI
    parse_args();

    // monitor the sonic server
    MY_TASK.run();
    
    // API
    let api_task = tokio::spawn(async move {
        run_api().await
    });

    let mut idle_task_storage = IMAP_TASK_STORAGE.lock().unwrap();
    idle_task_storage.start_downloading();
    drop(idle_task_storage);
    start_checkin_for_all_updates();
    
    // dont stop the program while API is running
    api_task.await.expect("Error while running API");
}

/// Entry point for a wasi env
#[cfg(target_os = "wasi")]
#[tokio::main(flavor = "current_thread")]
async fn main() -> anyhow::Result<()> {
    let start = Instant::now();
    
    run().await;
    
    // let email = LettreMessage::builder()
    //     .from("sokolovskiiyura@outlook.com".parse().unwrap())
    //     .to("sokolovskiiyura@gmail.com".parse().unwrap())
    //     .subject("Hello, World!")
    //     .multipart(
    //         MultiPart::alternative() // This is composed of two parts.
    //             .singlepart(
    //                 SinglePart::plain("Hello, World!".to_string()) // Plain text version
    //             )
    //             .singlepart(
    //                 SinglePart::html("<h1>Hello, World!</h1>".to_string()) // HTML version
    //             )
    //     )
    //     .unwrap();
    
    
    // let _ = send_email(email).await;
    
    
    // delete_email("Sent".to_string(), 4).await;
    
    println!("{:?}", Instant::now() - start);
    Ok(())
}

/// Entry point for a not wasi env
#[cfg(not(target_os = "wasi"))]
#[tokio::main]
async fn main() -> anyhow::Result<()>{
    let start = Instant::now();
    
    run().await;
    
    // let email = LettreMessage::builder()
    //     .from("sokolovskiiyura@outlook.com".parse().unwrap())
    //     .to("sokolovskiiyura@gmail.com".parse().unwrap())
    //     .subject("Hello, World!")
    //     .multipart(
    //         MultiPart::alternative() // This is composed of two parts.
    //             .singlepart(
    //                 SinglePart::plain("Hello, World!".to_string()) // Plain text version
    //             )
    //             .singlepart(
    //                 SinglePart::html("<h1>Hello, World!</h1>".to_string()) // HTML version
    //             )
    //     )
    //     .unwrap();
 
    
    // let _ = send_email(email).await;

    
    // delete_email("lqo7m8r2.1js7w080jvr2o@express.medallia.com".to_string()).await.expect("Unable to delete an email");

    println!("{:?}", Instant::now() - start);
    
    Ok(())
}
